from ._classic import A, B, C, D
from ._exceptionals import E, F4, G2
from ._base import LieAlgebra, NumericSymbol

__all__ = [
    "A", "B", "C", "D", "F4", "G2", "E", "LieAlgebra", "NumericSymbol"
]



"""
Liesym is an extension module on SymPy that reimplements the
liealgebra module inside sympy with the idea of using a
compiled backend to do speedups on some of the slower aspects
of SymPy's pure python implementation. This python module is
mostly a reimplementation of a Mathematica module LieART.


* SymPy: https://sympy.org
* LieART: https://arxiv.org/pdf/1206.6379.pdf
"""

# Compiled extension imports
from .liesym import (
    debug_mode as _debug_mode,
    LieAlgebraBackend as _LieAlgebraBackend,
)

from .algebras import *

__all__ = [
    "A", "B", "C", "D", "F4", "G2", "E", "LieAlgebra", "NumericSymbol"
]



import numpy as np
from sympy import Matrix, flatten, Rational

from .. import (_debug_mode, _LieAlgebraBackend)


def _to_rational_tuple(obj):
    if obj is None:
        return
    elif isinstance(obj, list) or isinstance(obj, tuple):
        obj = Matrix(obj)
    elif isinstance(obj, Rational):
        return int(obj)
    elif isinstance(obj, int) or isinstance(obj, np.ndarray):
        return obj
    else:
        pass
    x = flatten(obj)
    return np.array([(i.p, i.q) for i in x], dtype=np.int64).reshape(*obj.shape, 2)


def _rust_new(func):
    def inner(*args, **kwargs):
        cls = args[0]
        nargs = [_to_rational_tuple(x) for x in args[1:]]
        return func(cls, *nargs, **kwargs)
    return inner


def _rust_wrapper(func):
    """Wraps the rust methods to and from. Formats the calls
    to rust by turning either a 2d matrix to 3d matrix of (x,y) => (x,y,[numerator,denominator])
    for preserving rational numbers. Rust returns objects as a tuple of (numerator-matrix, denominator-matrix)
    """
    def inner(*args, **kwargs):
        cls = args[0]
        rank = cls.rank

        nargs = [_to_rational_tuple(x) for x in args[1:]]
        result = func(cls, *nargs, **kwargs)

        if isinstance(result, (int, str, float, bool)) or result is None:
            return result

        if not isinstance(result, tuple):
            return

        # Tuple of ints
        if isinstance(result[0], int) and isinstance(result[1], int):
            return Rational(*result)

        # tuple of ndarrays
        numer, denom = (x.squeeze() for x in result)
        shape = numer.shape

        plain_values = [Rational(f"{x}/{y}")
                        for x, y in zip(numer.flatten(), denom.flatten())]
        # vectorlike
        if len(shape) == 1:
            shape = (shape[0], 1) if rank == 1 else (1, shape[0])

        m = Matrix(*shape, plain_values)
        return [m.row(i) for i in range(m.shape[0])]
    return inner


class _LieAlgebraBackendWrapped:
    @_rust_new
    def __init__(self, *args, **kwargs):
        # obscuring this option
        backend = kwargs.get("backend", _LieAlgebraBackend)
        self.rank = args[0]
        self.backend = backend(*args)

    @_rust_wrapper
    def orbit(self, weight, stabilizers):
        return self.backend.orbit(weight, stabilizers)

    @_rust_wrapper
    def root_system(self):
        return self.backend.root_system()

    @_rust_wrapper
    def tensor_product_decomposition(self, irrepA, irrepB):
        return self.backend.tensor_product_decomposition(irrepA, irrepB)

    @_rust_wrapper
    def dim(self, irrep):
        return self.backend.dim(irrep)

    @_rust_wrapper
    def get_irrep_by_dim(self, dim, max_dd):
        return self.backend.irrep_by_dim(dim, max_dd)

    @_rust_wrapper
    def index_irrep(self, irrep, dim):
        return self.backend.index_irrep(irrep, dim)

    @_rust_wrapper
    def conjugate(self, irrep):
        return self.backend.conjugate_irrep(irrep)


def create_backend(algebra):
    return _LieAlgebraBackendWrapped(
        algebra.rank,
        algebra.n_pos_roots,
        algebra.simple_roots,
        algebra.cartan_matrix,
        algebra.cartan_matrix.pinv(),
        algebra.omega_matrix,
        algebra.omega_matrix.pinv(),
        algebra.cocartan_matrix,
    )



from sympy.core.sympify import _sympify
from sympy import Matrix, flatten

from ._base import LieAlgebra


def _euclidean_root(i, n):
    root = [0]*n
    root[i] = 1
    try:
        root[i+1] = -1
    except IndexError:
        pass  # catches B last root
    return Matrix([root])


class A(LieAlgebra):
    r"""The compact lie group of type A. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_A.png
           :width: 300px
           :align: center

    """
    def __new__(cls, n):
        return super().__new__(cls, "A", _sympify(n))

    def __init__(self, *args, **kwargs):
        super().__init__()
        n = self.rank
        self._simple_roots = [_euclidean_root(i, n+1) for i in range(n)]

    @property
    def dimension(self) -> int:
        """The dimension of the simple Lie algebra A series is
        one greater than the rank of the algebra.

        .. math::
            n + 1
        """
        return self.rank + 1

    @property
    def n_pos_roots(self) -> int:
        """The number of positive roots for the simple Lie algebra A is 
        defined as 

        .. math::
            n(n + 1) / 2
        """
        return self.rank * (self.rank + 1) / 2

    def dim_name(self, irrep: Matrix) -> str:
        r"""Returns a sympy formatted symbol for the irrep.
        This is commonly used in physics literature"""
        if self.rank == 2:
            if irrep == Matrix([[2, 0]]):
                return self._dim_name_fmt(6)
            elif irrep == Matrix([[0, 2]]):
                return self._dim_name_fmt(6, True)
        return super().dim_name(irrep)

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + int(len(l) < 5)

    def _congruency_class(self, irrep):
        r = self.rank+1
        v = int(
            Matrix([range(1, r)]).dot(irrep)
        )
        return v % r


class B(LieAlgebra):
    r"""The compact lie group of type B. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_B.png
           :width: 300px
           :align: center

    """
    def __new__(cls, n):
        return super().__new__(cls, "B", _sympify(n))

    def __init__(self, *args, **kwargs):
        super().__init__()
        n = self.rank
        self._simple_roots = [_euclidean_root(i, n) for i in range(n)]

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + int(len(l) < 5)

    @property
    def dimension(self) -> int:
        """The dimension of the simple Lie algebra B series is
        equal to the rank of the algebra.
        """

        return self.rank

    @property
    def n_pos_roots(self) -> int:
        """The number of positive roots for the simple Lie algebra B is 
        defined as 

        .. math::
            n^2
        """
        return self.rank**2

    def _congruency_class(self, irrep):
        return flatten(irrep.tolist())[-1] % 2


class C(LieAlgebra):
    r"""The compact lie group of type C. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_C.png
           :width: 300px
           :align: center
    """

    def __new__(cls, n):
        return super().__new__(cls, "C", _sympify(n))

    def __init__(self, *args, **kwargs):
        super().__init__()
        n = self.rank
        c_root = [0] * n
        c_root[-1] = 2
        self._simple_roots = [_euclidean_root(i, n) for i in range(
            n - 1)] + [Matrix([c_root])]

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + int(len(l) < 5)

    @property
    def dimension(self) -> int:
        """The dimension of the simple Lie algebra C series is
        equal to the rank of the algebra.
        """

        return self.rank

    @property
    def n_pos_roots(self) -> int:
        """The number of positive roots for the simple Lie algebra C is 
        defined as 

        .. math::
            n^2
        """

        return self.rank**2

    def _congruency_class(self, irrep):
        return sum(flatten(irrep.tolist())[::2]) % 2


class D(LieAlgebra):
    r"""The compact lie group of type D. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_D.png
           :width: 300px
           :align: center
    """

    def __new__(cls, n):
        if n < 2:
            raise ValueError("Min dimension for type 'D' is 2")
        return super().__new__(cls, "D", _sympify(n))

    def __init__(self, *args, **kwargs):
        super().__init__()
        n = self.rank
        d_root = [0] * n
        d_root[-1] = 1
        d_root[-2] = 1
        self._simple_roots = [_euclidean_root(i, n) for i in range(
            n - 1)] + [Matrix([d_root])]

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + int(len(l) < 5)

    @property
    def dimension(self) -> int:
        """The dimension of the simple Lie algebra D series is
        equal to the rank of the algebra.
        """

        return self.rank

    @property
    def n_pos_roots(self) -> int:
        """The number of roots for the simple Lie algebra D is 
        defined as 

        .. math::
            n(n-1)
        """

        return self.rank * (self.rank - 1)

    def _congruency_class(self, irrep):
        l = flatten(irrep.tolist())
        i = sum(l[-2:]) % 2
        n = self.rank
        j = sum(l[:-2][::2]) + (n-2) * l[-2] + n * l[-1]

        return (i, j)



from sympy import Matrix, S, flatten, sympify

from ._base import LieAlgebra


class F4(LieAlgebra):
    r"""The compact lie group of type F4. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_F4.png
           :height: 50px
           :align: center

    """

    def __new__(cls):
        return super().__new__(cls, "F", sympify(4))

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._simple_roots = [
            Matrix([[1, -1, 0, 0]]),
            Matrix([[0, 1, -1, 0]]),
            Matrix([[0, 0, 1, 0]]),
            Matrix([[-S.Half, -S.Half, -S.Half, -S.Half]]),
        ]

    @property
    def dimension(self) -> int:
        return 4

    @property
    def n_pos_roots(self) -> int:
        return 24

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + 1


class G2(LieAlgebra):
    r"""The compact lie group of type G2. The dynkin diagram for this algebra is

        .. image:: ../../docs/source/images/type_G2.png
           :height: 50px
           :align: center

    """
    def __new__(cls):
        return super().__new__(cls, "G", sympify(2))

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._simple_roots = [
            Matrix([[0, 1, -1]]),
            Matrix([[1, -2, 1]])
        ]

    @property
    def dimension(self) -> int:
        return 3

    @property
    def n_pos_roots(self) -> int:
        return 6

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep.tolist())
        return max(l) + 3


def _e_series_default_roots(n):
    e8 = [
        [S.Half, -S.Half, -S.Half,
         -S.Half, -S.Half, -S.Half,
         -S.Half, S.Half],
        [-1, 1, 0, 0, 0, 0, 0, 0],
        [0, -1, 1, 0, 0, 0, 0, 0],
        [0, 0, -1, 1, 0, 0, 0, 0],
        [0, 0, 0, -1, 1, 0, 0, 0],
        [0, 0, 0, 0, -1, 1, 0, 0],
        [0, 0, 0, 0, 0, -1, 1, 0],
        [1, 1, 0, 0, 0, 0, 0, 0]]

    roots = e8[:n - 1] + [e8[-1]]
    return [Matrix([roots[i]]) for i in range(n)]


class E(LieAlgebra):
    r"""The compact lie group of type E. There are only three defined for
    type E.


        .. figure:: ../../docs/source/images/type_E6.png
           :height: 100px
           :align: center

           E6


        .. figure:: ../../docs/source/images/type_E7.png
           :height: 100px
           :align: center

           E7


        .. figure:: ../../docs/source/images/type_E8.png
           :height: 100px
           :align: center

           E8

    """
    def __new__(cls, n):
        if n not in [6, 7, 8]:
            raise ValueError("Algebra series E only defined for 6, 7 and 8}")
        return super().__new__(cls, "E", sympify(n))

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._simple_roots = _e_series_default_roots(self.args[1])

    @property
    def dimension(self) -> int:
        return self.rank

    @property
    def n_pos_roots(self) -> int:
        return [36, 63, 120][self.rank-6]

    def max_dynkin_digit(self, irrep: Matrix) -> int:
        """Returns the max Dynkin Digit for the representation"""
        l = flatten(irrep).tolist()
        if self.rank == 6:
            return max(l) + 3
        return max(l)

    def _congruency_class(self, irrep):
        n = self.rank
        if n == 8:
            return 0

        l = flatten(irrep.tolist())
        if n == 7:
            return (l[3] + l[5] + l[6]) % 2

        if n == 6:
            return (l[0] - l[1] + l[3] - l[4]) % 3



from typing import List
from sympy import zeros, Matrix, eye, sqrt


def _cartan_matrix(simple_roots: List[Matrix]) -> Matrix:
    rank = len(simple_roots)
    cartan_matrix = zeros(rank, rank)
    for i, sr_i in enumerate(simple_roots):
        for j, sr_j in enumerate(simple_roots):
            cartan_matrix[j, i] = 2 * sr_i.dot(sr_j) / sr_i.dot(sr_i)
    return cartan_matrix


def _cocartan_matrix(simple_roots: List[Matrix]) -> Matrix:
    return Matrix([2 * x / x.dot(x) for x in simple_roots])


def _quadratic_form(cartan_matrix: Matrix, simple_roots: List[Matrix]) -> Matrix:
    rank = len(simple_roots)
    quadratic_form = zeros(rank, rank)

    # normalized constant
    n_constant = sqrt(2 / max(x.dot(x) for x in simple_roots))
    for i in range(rank):
        root = n_constant * simple_roots[i]
        quadratic_form[i, i] = root.dot(root) / 2

    return cartan_matrix.pinv() * quadratic_form


def _reflection_matricies(simple_roots: List[Matrix]) -> List[Matrix]:
    def reflection_matrix(v): return (
        eye(len(v)) - 2 * v.T * v / v.dot(v)).as_immutable()
    return [reflection_matrix(x) for x in simple_roots]



from setuptools import find_packages, setup
from setuptools_rust import RustExtension, Binding
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup_requires = [
    'setuptools-rust>=0.11.6', 
    'pytest-runner'
]

install_requires = [
    'numpy>=1.16',
    'sympy>=1.5',
]

setup(
    name='liesym',
    version='0.3.0',
    description='Sympy Lie Algebra extensions powered by rust.',
    rust_extensions=[RustExtension(
        'liesym.liesym',
        './Cargo.toml',
        binding=Binding.PyO3,
        debug=False
    )],
    install_requires=install_requires,
    setup_requires=setup_requires,
    python_requires=">=3.7",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    test_requires=["pytest"],
    author="Nathan Papapietro <npapapietro95@gmail.com>",
    url="https://github.com/npapapietro/liesym",
    classifiers=[
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Rust",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown'
)
